from typing import Any

from verse.content.image import Image, ImageParam
from verse.core import Component, Response

from ._operation import ZeroShotImageClassificationOperation


class ZeroShotImageClassification(Component):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def classify(
        self,
        image: str | bytes | Image | ImageParam | dict,
        classes: list[str],
        **kwargs: Any,
    ) -> Response[list[float]]:
        """Zero-shot classification of image with provided texts.

        Args:
            image: Image to classify.
            texts: List of classes as texts.

        Returns:
            Classification results.
        """
        return self._run_internal(
            ZeroShotImageClassificationOperation.CLASSIFY, locals()
        )

    def batch(
        self,
        images: list[str | bytes | Image | ImageParam | dict],
        classes: list[str],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Zero-shot classification of images with provided texts.

        Args:
            images: Images to classify.
            texts: List of classes as texts.

        Returns:
            List of classification results.
        """
        return self._run_internal(
            ZeroShotImageClassificationOperation.BATCH, locals()
        )

    async def aclassify(
        self,
        image: Image,
        classes: list[str],
        **kwargs: Any,
    ) -> Response[list[float]]:
        """Zero-shot classification of image with provided texts.

        Args:
            image: Image to classify.
            texts: List of classes as texts.

        Returns:
            Classification results.
        """
        return await self._arun_internal(
            ZeroShotImageClassificationOperation.CLASSIFY, locals()
        )

    async def abatch(
        self,
        images: list[Image],
        classes: list[str],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Zero-shot classification of images with provided texts.

        Args:
            images: Images to classify.
            texts: List of classes as texts.

        Returns:
            List of classification results.
        """
        return await self._arun_internal(
            ZeroShotImageClassificationOperation.BATCH, locals()
        )



from typing import Any

from verse.content.image import Image, ImageParam
from verse.core import Component, Response

from ._operation import ImageCaptioningOperation


class ImageCaptioning(Component):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def generate(
        self,
        image: str | bytes | Image | ImageParam | dict,
        max_length: int | None = None,
        **kwargs: Any,
    ) -> Response[str]:
        """Generate caption for image.

        Args:
            image: Image.

        Returns:
            Caption.
        """
        return self._run_internal(ImageCaptioningOperation.GENERATE, locals())

    def batch(
        self,
        images: list[str | bytes | Image | ImageParam | dict],
        max_length: int | None = None,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """Generate captions for batch of images.

        Args:
            images: List of images.

        Returns:
            List of captions.
        """
        return self._run_internal(ImageCaptioningOperation.BATCH, locals())

    async def agenerate(
        self,
        image: str | bytes | Image | ImageParam | dict,
        **kwargs: Any,
    ) -> Response[str]:
        """Generate caption for image.

        Args:
            image: Image.

        Returns:
            Caption.
        """
        return await self._arun_internal(
            ImageCaptioningOperation.GENERATE, locals()
        )

    async def abatch(
        self,
        images: list[str | bytes | Image | ImageParam | dict],
        **kwargs: Any,
    ) -> Response[list[str]]:
        """Generate captions for batch of images.

        Args:
            images: List of images.

        Returns:
            List of captions.
        """
        return await self._arun_internal(
            ImageCaptioningOperation.BATCH, locals()
        )



from typing import Any

from verse.content.image import Image, ImageParam
from verse.core import Component, Response

from ._operation import ImageEmbeddingOperation


class ImageEmbedding(Component):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def embed(
        self,
        image: str | bytes | Image | ImageParam | dict,
        **kwargs: Any,
    ) -> Response[list[float]]:
        """Create embedding for image.

        Args:
            image: Image.

        Returns:
            Image embedding.
        """
        return self._run_internal(ImageEmbeddingOperation.EMBED, locals())

    def compare(
        self,
        embedding1: list[float],
        embedding2: list[float],
        **kwargs,
    ) -> Response[float]:
        """Compare embeddings using cosine similarity.

        Args:
            embedding1: Embedding 1.
            embedding2: Embedding 2.

        Returns:
            Similarity score.
        """
        return self._run_internal(ImageEmbeddingOperation.COMPARE, locals())

    def batch(
        self,
        images: list[str | bytes | Image | ImageParam | dict],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Create embedding for batch of images.

        Args:
            images: List of images.

        Returns:
            List of image embeddings.
        """
        return self._run_internal(ImageEmbeddingOperation.BATCH, locals())

    def embed_text(
        self,
        text: str,
        **kwargs: Any,
    ) -> Response[list[float]]:
        """Create embedding for text.

        Args:
            text: Text string.

        Returns:
            Text embedding.
        """
        return self._run_internal(ImageEmbeddingOperation.EMBED_TEXT, locals())

    def batch_text(
        self,
        texts: list[str],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Create embedding for batch of text.

        Args:
            texts: List of texts.

        Returns:
            List of text embeddings.
        """
        return self._run_internal(ImageEmbeddingOperation.BATCH_TEXT, locals())

    def batch_image_text(
        self,
        images: list[str | bytes | Image | ImageParam | dict],
        texts: list[str],
        **kwargs: Any,
    ) -> Response[tuple[list[list[float]], list[list[float]]]]:
        """Create embedding for batch of images and texts.

        Args:
            images: List of images.
            texts: List of texts.

        Returns:
            Tuple of list of image embeddings and text embeddings.
        """
        return self._run_internal(
            ImageEmbeddingOperation.BATCH_IMAGE_TEXT, locals()
        )

    async def aembed(
        self,
        image: Image | ImageParam | dict,
        **kwargs: Any,
    ) -> Response[list[float]]:
        """Create embedding for image.

        Args:
            image: Image.

        Returns:
            Image embedding.
        """
        return await self._arun_internal(
            ImageEmbeddingOperation.EMBED, locals()
        )

    async def acompare(
        self,
        embedding1: list[float],
        embedding2: list[float],
        **kwargs,
    ) -> Response[float]:
        """Compare embeddings using cosine similarity.

        Args:
            embedding1: Embedding 1.
            embedding2: Embedding 2.

        Returns:
            Similarity score.
        """
        return await self._arun_internal(
            ImageEmbeddingOperation.COMPARE, locals()
        )

    async def abatch(
        self,
        images: list[Image | ImageParam | dict],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Create embedding for batch of images.

        Args:
            images: List of images.

        Returns:
            List of image embeddings.
        """
        return await self._arun_internal(
            ImageEmbeddingOperation.BATCH, locals()
        )

    async def aembed_text(
        self,
        text: str,
        **kwargs: Any,
    ) -> Response[list[float]]:
        """Create embedding for text.

        Args:
            text: Text string.

        Returns:
            Text embedding.
        """
        return await self._arun_internal(
            ImageEmbeddingOperation.EMBED_TEXT, locals()
        )

    async def abatch_text(
        self,
        texts: list[str],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Create embedding for batch of text.

        Args:
            texts: List of texts.

        Returns:
            List of text embeddings.
        """
        return await self._arun_internal(
            ImageEmbeddingOperation.BATCH_TEXT, locals()
        )

    async def abatch_image_text(
        self,
        images: list[Image | ImageParam | dict],
        texts: list[str],
        **kwargs: Any,
    ) -> Response[tuple[list[list[float]], list[list[float]]]]:
        """Create embedding for batch of images and texts.

        Args:
            images: List of images.
            texts: List of texts.

        Returns:
            Tuple of list of image embeddings and text embeddings.
        """
        return await self._arun_internal(
            ImageEmbeddingOperation.BATCH_IMAGE_TEXT, locals()
        )



"""
Speech recognition using Azure Speech.
"""

__all__ = ["AzureSpeech"]

import json
from typing import Any

import azure.cognitiveservices.speech as speechsdk
from scipy.io.wavfile import write

from verse.content.audio import Audio
from verse.core import Context, Operation, Provider, Response

from .._models import SpeechRecognitionResult
from .._operation import SpeechRecognitionOperation
from .._operation_parser import SpeechRecognitionOperationParser


class AzureSpeech(Provider):
    api_key: str
    service_region: str

    def __init__(
        self,
        api_key: str = "",
        service_region: str = "",
        languages: list[str] = ["en-US", "en-GB", "es-ES", "fr-FR"],
        **kwargs,
    ):
        """Intialize.

        Args:
            api_key:
                Azure speech api key.
            service_region:
                Azure speech service region.
        """

        self.speech_config = speechsdk.SpeechConfig(
            subscription=api_key, region=service_region
        )
        self.speech_config.output_format = speechsdk.OutputFormat.Detailed
        self.auto_detect_source_language_config = (
            speechsdk.languageconfig.AutoDetectSourceLanguageConfig(
                languages=languages
            )
        )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.init(context=context)
        op_parser = SpeechRecognitionOperationParser(operation)
        result: Any = None
        # TRANSCRIBE
        if op_parser.op_equals(SpeechRecognitionOperation.TRANSCRIBE):
            result = self._transcribe(audio=op_parser.get_audio())
        # DETECT_LANGUAGE
        elif op_parser.op_equals(SpeechRecognitionOperation.DETECT_LANGUAGE):
            result = self._detect_language(audio=op_parser.get_audio())
        else:
            return super().run(
                operation,
                context,
                **kwargs,
            )
        return Response(result=result)

    def _transcribe(self, audio: Audio):
        audio = audio.convert(
            type="bytes",
            format="wav",
            codec="pcm_s16le",
            rate=16000,
            channels=1,
        )
        audio_stream_callback = AudioInputStreamCallback(audio)

        stream = speechsdk.audio.PullAudioInputStream(
            pull_stream_callback=audio_stream_callback
        )
        audio_input = speechsdk.AudioConfig(stream=stream)
        speech_recognizer = speechsdk.SpeechRecognizer(
            speech_config=self.speech_config, audio_config=audio_input
        )
        result = speech_recognizer.recognize_once()
        json_obj = json.loads(result.json)

        return SpeechRecognitionResult(
            text=json_obj["DisplayText"], verbose=json_obj, language=None
        )

    def _detect_language(self, audio: Audio):
        audio = audio.convert(
            type="bytes",
            format="wav",
            codec="pcm_s16le",
            rate=16000,
            channels=1,
        )
        audio_stream_callback = AudioInputStreamCallback(audio)

        stream = speechsdk.audio.PullAudioInputStream(
            pull_stream_callback=audio_stream_callback
        )
        audio_input = speechsdk.AudioConfig(stream=stream)

        speech_recognizer = speechsdk.SpeechRecognizer(
            speech_config=self.speech_config,
            auto_detect_source_language_config=self.auto_detect_source_language_config,
            audio_config=audio_input,
        )
        result = speech_recognizer.recognize_once()
        detected_language = result.properties[
            speechsdk.PropertyId.SpeechServiceConnection_AutoDetectSourceLanguageResult
        ]
        json_obj = json.loads(result.json)
        return detected_language

    def _extract_and_resample_audio(self, stream, target_rate=16000):
        pass


class AudioInputStreamCallback(speechsdk.audio.PullAudioInputStreamCallback):
    def __init__(self, audio_data):
        super().__init__()
        self.audio_data = audio_data
        self.read_pos = 0  # Position in the audio data

    def read(self, buffer):
        # Read data into the buffer
        bytes_requested = len(buffer)
        bytes_available = len(self.audio_data) - self.read_pos
        bytes_to_copy = min(bytes_requested, bytes_available)

        if bytes_to_copy > 0:
            buffer[:bytes_to_copy] = self.audio_data[
                self.read_pos : self.read_pos + bytes_to_copy
            ]
            self.read_pos += bytes_to_copy
            return bytes_to_copy
        else:
            return 0  # No more data to read

    def close(self):
        pass



"""
Speech recognition using Whisper.
"""

__all__ = ["Whisper"]

from typing import Any

import av
import numpy as np
import whisper
from whisper import Whisper as OpenAIWhisper

from verse.content.audio import Audio
from verse.core import Context, Operation, Provider, Response

from .._operation import SpeechRecognitionOperation
from .._operation_parser import SpeechRecognitionOperationParser

from .._models import SpeechRecognitionResult


class Whisper(Provider):
    model: str
    device: str | None

    _model: OpenAIWhisper

    def __init__(
        self,
        model: str = "tiny",
        device: str | None = None,
        **kwargs,
    ):
        """Intialize.

        Args:
            model:
                Whisper model name.
            device:
                Pytorch device name.
        """
        self.model = model
        self.device = device
        self._model = None

    def init(self, context: Context | None = None) -> None:
        if self._model:
            return
        self._model = whisper.load_model(name=self.model, device=self.device)

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.init(context=context)
        op_parser = SpeechRecognitionOperationParser(operation)
        result: Any = None
        # TRANSCRIBE
        if op_parser.op_equals(SpeechRecognitionOperation.TRANSCRIBE):
            result = self._transcribe(audio=op_parser.get_audio())
        # DETECT_LANGUAGE
        elif op_parser.op_equals(SpeechRecognitionOperation.DETECT_LANGUAGE):
            result = self._detect_language(audio=op_parser.get_audio())
        else:
            return super().run(
                operation,
                context,
                **kwargs,
            )
        return Response(result=result)

    def _transcribe(self, audio: Audio):
        waudio = self._extract_and_resample_audio(audio.convert(type="stream"))
        waudio = whisper.pad_or_trim(waudio)
        result = whisper.transcribe(self._model, waudio)
        result = SpeechRecognitionResult(text=result["text"], 
                                        language=result["language"],
                                        verbose=result)
        return result

    def _detect_language(self, audio: Audio):
        whisper_audio = self._extract_and_resample_audio(audio.convert(type="stream"))
        whisper_audio = whisper.pad_or_trim(whisper_audio)
        mel = whisper.log_mel_spectrogram(whisper_audio).to(self._model.device)
        _, probs = self._model.detect_language(mel)
        return max(probs, key=probs.get)

    def _extract_and_resample_audio(self, stream, target_rate=16000):
        container = av.open(stream)
        stream = container.streams.audio[0]

        resampler = av.audio.resampler.AudioResampler(
            format="s16", layout="mono", rate=target_rate
        )

        samples = []

        for frame in container.decode(stream):
            resampled_frames = resampler.resample(frame)
            for resampled_frame in resampled_frames:
                samples.append(resampled_frame.to_ndarray())

        audio_data = np.concatenate(samples, axis=1)
        if audio_data.ndim > 1:
            audio_data = audio_data.mean(axis=0)

        audio_data = audio_data / np.max(np.abs(audio_data))
        if audio_data.dtype != np.float32:
            audio_data = audio_data.astype(np.float32)

        return audio_data



from __future__ import annotations

from typing import Any

from verse.core import Component, Response

from ._models import (
    ChatCompletionMessage,
    ChatCompletionResult,
    ResponseFormat,
    Stream,
    StreamOptions,
    Tool,
    ToolChoice,
)
from ._operation import ChatCompletionOperation


class ChatCompletion(Component):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def complete(
        self,
        messages: str | list[dict | ChatCompletionMessage],
        model: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_tokens: int | None = None,
        stop: str | list[str] | None = None,
        n: int | None = None,
        presence_penalty: float | None = None,
        frequency_penalty: float | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: str | dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        response_format: dict | ResponseFormat | None = None,
        seed: int | None = None,
        **kwargs: Any,
    ) -> Response[ChatCompletionResult]:
        return self._run_internal(ChatCompletionOperation.COMPLETE, locals())

    def stream(
        self,
        messages: str | list[dict | ChatCompletionMessage],
        model: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_tokens: int | None = None,
        stop: str | list[str] | None = None,
        n: int | None = None,
        presence_penalty: float | None = None,
        frequency_penalty: float | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: str | dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        response_format: dict | ResponseFormat | None = None,
        seed: int | None = None,
        stream_options: StreamOptions | None = None,
        **kwargs: Any,
    ) -> Response[Stream[ChatCompletionResult]]:
        return self._run_internal(ChatCompletionOperation.STREAM, locals())

    async def acomplete(
        self,
        messages: str | list[dict | ChatCompletionMessage],
        model: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_tokens: int | None = None,
        stop: str | list[str] | None = None,
        n: int | None = None,
        presence_penalty: float | None = None,
        frequency_penalty: float | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: str | dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        response_format: dict | ResponseFormat | None = None,
        seed: int | None = None,
        **kwargs: Any,
    ) -> Response[ChatCompletionResult]:
        return await self._arun_internal(
            ChatCompletionOperation.COMPLETE, locals()
        )

    async def astream(
        self,
        messages: str | list[dict | ChatCompletionMessage],
        model: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_tokens: int | None = None,
        stop: str | list[str] | None = None,
        n: int | None = None,
        presence_penalty: float | None = None,
        frequency_penalty: float | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: str | dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        response_format: dict | ResponseFormat | None = None,
        seed: int | None = None,
        stream_options: StreamOptions | None = None,
        **kwargs: Any,
    ) -> Response[Stream[ChatCompletionResult]]:
        return await self._arun_internal(
            ChatCompletionOperation.STREAM, locals()
        )



"""
Speech Recognition using Google Speech-to-Text.
"""

__all__ = ["GoogleSpeechToText"]
import json
from typing import Any
from verse.core import Provider
from google.cloud import speech_v1p1beta1 as speech
from google.oauth2 import service_account

from verse.content.audio import Audio
from verse.core import Context, Operation, Provider, Response

from .._operation import SpeechRecognitionOperation
from .._operation_parser import SpeechRecognitionOperationParser
from .._models import SpeechRecognitionResult

class GoogleSpeechToText(Provider):
    def __init__(
        self,
        api_key_path: str = "",
        languages: list[str] = ["en-US", "en-GB", "es-ES", "fr-FR"],
        **kwargs,
    ):
        """Intialize.

        Args:
            api_key:
                Azure speech api key.
            service_region:
                Azure speech service region.
        """
        credentials = service_account.Credentials.from_service_account_file(api_key_path)
        self.client = speech.SpeechClient(credentials=credentials)
        self.languages = languages
        self.config = speech.RecognitionConfig(
            encoding=speech.RecognitionConfig.AudioEncoding.ENCODING_UNSPECIFIED,
            sample_rate_hertz=16000,
            language_code="en-US",  # Fallback language
            alternative_language_codes=languages,  # Possible languages
        )
    
    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = SpeechRecognitionOperationParser(operation)
        result: Any = None
        # TRANSCRIBE
        if op_parser.op_equals(SpeechRecognitionOperation.TRANSCRIBE):
            result = self._transcribe(audio=op_parser.get_audio())
        # DETECT_LANGUAGE
        elif op_parser.op_equals(SpeechRecognitionOperation.DETECT_LANGUAGE):
            result = self._detect_language(audio=op_parser.get_audio())
        else:
            return super().run(operation)
        return Response(result=result)

    def _transcribe(self, audio: Audio):
        audio = audio.convert(type="bytes", format="wav", codec="pcm_s16le", rate=16000, channels=1)
        audio = speech.RecognitionAudio(content=audio)
        
        response = self.client.recognize(config=self.config, audio=audio)
        
        return SpeechRecognitionResult(
            text=response.results[0].alternatives[0].transcript,
            verbose=response,
            language=response.results[0].language_code,
        )

    def _detect_language(self, audio: Audio):
        
        return self._transcribe(audio).language


from __future__ import annotations

__all__ = ["Cohere"]

import json
import uuid
from typing import Any

from cohere import NonStreamedChatResponse, StreamedChatResponse
from cohere import ToolCall as CohereToolCall
from cohere import ToolResult as CohereToolResult

from verse.core import Context, NCall, Operation, Provider, Response
from verse.core.exceptions import BadRequestError

from .._helper import copy_arg
from .._models import (
    ChatCompletionMessage,
    ChatCompletionMessageContentPart,
    ChatCompletionResult,
    Choice,
    Stream,
    Tool,
    ToolCall,
    ToolCallFunction,
    Usage,
)
from .._operation import ChatCompletionOperation
from .._operation_parser import ChatCompletionOperationParser


class Cohere(Provider):
    api_key: str | None
    model: str | None
    base_url: str | None
    client_name: str | None
    timeout: float | None
    raw_prompting: bool | None
    force_single_step: bool | None
    system_message_in_preamble: bool
    nparams: dict

    _client: Any
    _aclient: Any

    def __init__(
        self,
        api_key: str | None = None,
        model: str = "command-r-plus",
        base_url: str | None = None,
        client_name: str | None = None,
        timout: float | None = None,
        raw_prompting: bool | None = None,
        force_single_step: bool | None = None,
        system_message_in_preamble: bool = True,
        nparams: dict = dict(),
    ):
        """Initialize.

        Args:
            api_key:
                Cohere API key.
            model:
                Cohere model.
            base_url:
                Cohere base url.
            client_name:
                Cohere client name.
            timout:
                Timeout for client.
            raw_prompting:
                Whether raw prompting is enabled in every request.
            force_single_step:
                Whether single step is enabled in every request.
            system_message_in_preamble:
                Whether system message should be sent as preamble
                or sent through chat history. Defaults to True.
            nparams:
                Native params for Cohere client.
        """
        self.model = model
        self.api_key = api_key
        self.base_url = base_url
        self.client_name = client_name
        self.timeout = timout
        self.raw_prompting = raw_prompting
        self.force_single_step = force_single_step
        self.system_message_in_preamble = system_message_in_preamble
        self.nparams = nparams

        self._client = None
        self._aclient = None

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        from cohere import Client

        self._client = Client(**self.get_client_params())

    async def ainit(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        from cohere import AsyncClient

        self._aclient = AsyncClient(**self.get_client_params())

    def get_client_params(self) -> dict:
        args: dict = {}
        if self.api_key is not None:
            args["api_key"] = self.api_key
        if self.base_url is not None:
            args["base_url"] = self.base_url
        if self.client_name is not None:
            args["client_name"] = self.client_name
        if self.timeout is not None:
            args["timeout"] = self.timeout
        args = args | self.nparams
        return args

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.init(context=context)
        op_parser = ChatCompletionOperationParser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            self._client,
            ClientHelper(self._client),
        )
        if ncall is None:
            return super().run(operation=operation, context=context, **kwargs)
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(call=ncall, result=nresult))

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.ainit(context=context)
        op_parser = ChatCompletionOperationParser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            self._aclient,
            AsyncClientHelper(self._aclient),
        )
        if ncall is None:
            return await super().arun(
                operation=operation, context=context, **kwargs
            )
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(call=ncall, result=nresult))

    def _get_ncall(
        self,
        op_parser: ChatCompletionOperationParser,
        client: Any,
        client_helper: Any,
    ) -> tuple[NCall | None, dict]:
        call = None
        state: dict = {}
        nargs = op_parser.get_nargs()
        op_converter = OperationConverter(
            self.model,
            self.raw_prompting,
            self.force_single_step,
            self.system_message_in_preamble,
        )
        # COMPLETE
        if op_parser.op_equals(ChatCompletionOperation.COMPLETE):
            args, state = op_converter.convert_complete(
                op_parser.get_args(strong_types=True)
            )
            call = NCall(
                client.chat,
                args,
                nargs,
            )
        # STREAM
        elif op_parser.op_equals(ChatCompletionOperation.STREAM):
            args, state = op_converter.convert_stream(
                op_parser.get_args(strong_types=True)
            )
            call = NCall(
                client_helper.stream,
                {"args": args, "nargs": nargs},
            )
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict,
        op_parser: ChatCompletionOperationParser,
    ) -> Any:
        result: Any = None
        result_converter = ResultConverter()
        # COMPLETE
        if op_parser.op_equals(ChatCompletionOperation.COMPLETE):
            result = result_converter.convert_complete_result(
                nresult,
                state,
            )
        # STREAM
        if op_parser.op_equals(ChatCompletionOperation.STREAM):
            result = result_converter.convert_stream_result(
                nresult,
                state,
            )
        return result


class ClientHelper:
    client: Any

    def __init__(self, client: Any):
        self.client = client

    def stream(self, args: dict, nargs: Any):
        return NCall(
            self.client.chat_stream,
            args,
            nargs,
        ).invoke()


class AsyncClientHelper:
    client: Any

    def __init__(self, client: Any):
        self.client = client

    async def stream(self, args: dict, nargs: Any):
        return NCall(
            self.client.chat_stream,
            args,
            nargs,
        ).invoke()


class ResultConverter:
    _generation_id: str
    _model: str

    def __init__(self):
        pass

    def convert_complete_result(
        self,
        nresult: NonStreamedChatResponse,
        state: dict,
    ) -> ChatCompletionResult:
        choice = Choice(
            finish_reason=self._convert_finish_reason(nresult.finish_reason),
            index=0,
            message=ChatCompletionMessage(
                role="assistant",
                content=nresult.text,
                tool_calls=self._convert_tool_calls(nresult.tool_calls),
            ),
        )
        return ChatCompletionResult(
            id=nresult.generation_id,
            choices=[choice],
            usage=self._convert_usage(nresult),
            model=state["model"],
        )

    def convert_stream_result(
        self,
        nresult: Any,
        state: dict,
    ) -> Stream:
        self._model = state["model"]
        return Stream(nresult, self.convert_stream_chunk)

    def convert_stream_chunk(
        self, chunk: StreamedChatResponse
    ) -> ChatCompletionResult:
        if chunk.event_type == "stream-start":
            self._generation_id = chunk.generation_id
            result = ChatCompletionResult(
                id=self._generation_id,
                choices=[
                    Choice(
                        index=0, delta=ChatCompletionMessage(role="assistant")
                    )
                ],
                model=self._model,
            )
        elif chunk.event_type == "text-generation":
            result = ChatCompletionResult(
                id=self._generation_id,
                choices=[
                    Choice(
                        index=0,
                        delta=ChatCompletionMessage(
                            role="assistant", content=chunk.text
                        ),
                    )
                ],
                model=self._model,
            )
        elif chunk.event_type == "tool-calls-generation":
            result = ChatCompletionResult(
                id=self._generation_id,
                choices=[
                    Choice(
                        index=0,
                        delta=ChatCompletionMessage(
                            role="assistant",
                            content=chunk.text,
                            tool_calls=self._convert_tool_calls(
                                chunk.tool_calls
                            ),
                        ),
                    )
                ],
                model=self._model,
            )
        elif chunk.event_type == "stream-end":
            result = ChatCompletionResult(
                id=self._generation_id,
                choices=[
                    Choice(
                        finish_reason=self._convert_finish_reason(
                            chunk.finish_reason
                        ),
                        index=0,
                        delta=ChatCompletionMessage(),
                    )
                ],
                model=self._model,
                usage=self._convert_usage(chunk.response),
            )
        return result

    def _convert_tool_calls(
        self, tool_calls: list[CohereToolCall] | None
    ) -> list[ToolCall] | None:
        if tool_calls is None:
            return None
        rtool_calls = []
        for tool_call in tool_calls:
            rtool_calls.append(
                ToolCall(
                    id=str(uuid.uuid4()),
                    function=ToolCallFunction(
                        name=tool_call.name,
                        arguments=json.dumps(tool_call.parameters),
                    ),
                )
            )
        return rtool_calls

    def _convert_finish_reason(self, finish_reason: str | None) -> str | None:
        finish_reason_map = {
            "COMPLETE": "stop",
            "STOP_SEQUENCE": "stop",
            "MAX_TOKENS": "length",
            "ERROR_TOXIC": "content_filter",
            "ERROR": "error",
            "ERROR_LIMIT": "error",
            "USER_CANCEL": "cancel",
            None: None,
        }
        return finish_reason_map[finish_reason]

    def _convert_usage(self, nresult: NonStreamedChatResponse) -> Usage | None:
        usage = None
        if (
            nresult.meta is not None
            and nresult.meta.billed_units is not None
            and nresult.meta.billed_units.input_tokens is not None
            and nresult.meta.billed_units.output_tokens is not None
        ):
            usage = Usage(
                completion_tokens=int(nresult.meta.billed_units.output_tokens),
                prompt_tokens=int(nresult.meta.billed_units.input_tokens),
                total_tokens=int(
                    nresult.meta.billed_units.input_tokens
                    + nresult.meta.billed_units.output_tokens
                ),
            )
        return usage


class OperationConverter:
    model: str | None
    raw_prompting: bool | None
    force_single_step: bool | None
    system_message_in_preamble: bool

    def __init__(
        self,
        model: str | None,
        raw_prompting: bool | None,
        force_single_step: bool | None,
        system_message_in_preamble: bool,
    ):
        self.model = model
        self.raw_prompting = raw_prompting
        self.force_single_step = force_single_step
        self.system_message_in_preamble = system_message_in_preamble

    def convert_complete(
        self,
        args: dict,
    ) -> tuple[dict, dict]:
        return self._convert_args(args)

    def convert_stream(
        self,
        args: dict,
    ) -> tuple[dict, dict]:
        return self._convert_args(args)

    def _convert_args(
        self,
        args: dict,
    ) -> tuple[dict, dict]:
        state: dict = {}
        role_map: dict = {
            "user": "USER",
            "assistant": "CHATBOT",
            "system": "SYSTEM",
            "tool": "TOOL",
        }
        rargs: dict = {}
        if "model" not in args or args["model"] is None:
            rargs["model"] = self.model
        else:
            copy_arg(args, rargs, "model")
        state["model"] = rargs["model"]
        chat_history: list = []
        tool_calls = []
        single_step_tool_results = []
        messages: list[ChatCompletionMessage] = args["messages"]
        for message in messages:
            chat_message: dict = {}
            chat_message["role"] = role_map[message.role]
            if message.tool_calls:
                tool_calls.extend(message.tool_calls)
                chat_message["tool_calls"] = self._convert_tool_calls(
                    message.tool_calls
                )
            if isinstance(message.content, str):
                chat_message["text"] = message.content
            elif isinstance(message.content, list):
                parts: list[ChatCompletionMessageContentPart] = message.content
                text = ""
                for part in parts:
                    if part.type == "text":
                        text = f"{text} {part.text}"
                    else:
                        raise BadRequestError(
                            "Chat message type not supported"
                        )
                chat_message["text"] = text
            if (
                chat_message["role"] == "SYSTEM"
                and self.system_message_in_preamble
            ):
                rargs["preamble"] = chat_message["text"]
            elif (
                chat_message["role"] == "CHATBOT"
                and message.tool_calls is not None
                and self.force_single_step
            ):
                pass
            elif chat_message["role"] == "TOOL":
                tool_results = []
                for tool_call in tool_calls:
                    if tool_call.id == message.tool_call_id:
                        outputs: list = []
                        if isinstance(message.content, str):
                            try:
                                json_object = json.loads(message.content)
                                if isinstance(json_object, dict):
                                    outputs.append(json_object)
                                elif isinstance(json_object, list):
                                    outputs.extend(json_object)
                            except ValueError:
                                outputs.append({"result": message.content})
                        else:
                            raise BadRequestError(
                                "Tool output should be encoded as a string"
                            )
                        tool_results.append(
                            CohereToolResult(
                                call=self._convert_tool_call(tool_call),
                                outputs=outputs,
                            )
                        )
                if not self.force_single_step:
                    chat_message["tool_results"] = tool_results
                    chat_history.append(chat_message)
                else:
                    single_step_tool_results = tool_results
            else:
                chat_history.append(chat_message)
        last_message = chat_history[-1]
        if last_message["role"] == "USER":
            rargs["message"] = last_message["text"]
            if len(chat_history) >= 2:
                rargs["chat_history"] = chat_history[:-1]
        elif last_message["role"] == "TOOL":
            # It will come here only for multi-step.
            rargs["tool_results"] = chat_message["tool_results"]
            rargs["message"] = ""
            if len(chat_history) >= 2:
                rargs["chat_history"] = chat_history[:-1]
        else:
            raise BadRequestError("Last message should be user or tool role")
        if len(chat_history) >= 2:
            rargs["chat_history"] = chat_history[:-1]
        copy_arg(args, rargs, "temperature")
        copy_arg(args, rargs, "max_tokens")
        copy_arg(args, rargs, "seed")
        copy_arg(args, rargs, "presence_penalty")
        copy_arg(args, rargs, "frequency_penalty")
        copy_arg(args, rargs, "top_p", "p")
        copy_arg(args, rargs, "top_k", "k")
        if "stop" in args:
            rargs["stop_sequences"] = (
                [args["stop"]]
                if isinstance(args["stop"], str)
                else args["stop"]
            )
        if "response_format" in args:
            rargs["response_format"] = {"type": args["response_format"].type}
        if "tools" in args:
            rargs["tools"] = self._convert_tools(args["tools"])
        if len(single_step_tool_results) > 0:
            rargs["tool_results"] = single_step_tool_results
        if self.raw_prompting:
            rargs["raw_prompting"] = self.raw_prompting
        if self.force_single_step:
            rargs["force_single_step"] = self.force_single_step
        copy_arg(args, rargs, "force_single_step", "force_single_step")
        request_options = self._convert_request_options(args)
        if request_options:
            args["request_options"] = request_options
        return rargs, state

    def _convert_request_options(self, args: dict) -> dict | None:
        request_options = {}
        if "timeout" in args:
            request_options["timeout_in_seconds"] = int(args["timeout"])
        if "extra_headers" in args:
            copy_arg(
                args, request_options, "extra_headers", "additional_headers"
            )
        if "extra_query" in args:
            copy_arg(
                args,
                request_options,
                "extra_query",
                "additional_query_parameters",
            )
        if "extra_body" in args:
            copy_arg(
                args,
                request_options,
                "extra_body",
                "additional_body_parameters",
            )
        if len(request_options) == 0:
            return None
        return request_options

    def _convert_tool_calls(
        self, tool_calls: list[ToolCall] | None
    ) -> list[CohereToolCall] | None:
        rtool_calls = []
        if tool_calls is None:
            return None
        for tool_call in tool_calls:
            rtool_calls.append(self._convert_tool_call(tool_call))
        return rtool_calls

    def _convert_tool_call(self, tool_call: ToolCall) -> CohereToolCall:
        return CohereToolCall(
            name=tool_call.function.name,
            parameters=(
                json.loads(tool_call.function.arguments)
                if tool_call.function.arguments is not None
                else {}
            ),
        )

    def _convert_tools(self, tools: list[Tool]) -> list:
        rtools = []
        for tool in tools:
            if tool.type == "function":
                ctool: dict[str, Any] = {}
                ctool["name"] = tool.function.name
                ctool["description"] = tool.function.description
                if (
                    tool.function.parameters is not None
                    and "type" in tool.function.parameters
                    and tool.function.parameters["type"] == "object"
                    and "properties" in tool.function.parameters
                ):
                    params: dict[str, Any] = {}
                    for name, property in tool.function.parameters[
                        "properties"
                    ].items():
                        params[name] = {}
                        params[name]["type"] = property["type"]
                        params[name]["description"] = property["description"]
                        if "enum" in property:
                            params[name]["description"] = (
                                params[name]["description"]
                                + " Possible enum values "
                                + ", ".join(property["enum"])
                                + "."
                            )
                        if "required" in tool.function.parameters:
                            if name in tool.function.parameters["required"]:
                                params[name]["required"] = True
                            else:
                                params[name]["required"] = False
                    ctool["parameter_definitions"] = params
                rtools.append(ctool)
        return rtools



"""
Zero shot image classification using CLIP model.
"""

__all__ = ["Clip"]

from typing import Any

import torch
from transformers import CLIPModel, CLIPProcessor

from verse.core import Context, Operation, Provider, Response

from .._operation import ZeroShotImageClassificationOperation
from .._operation_parser import ZeroShotImageClassificationOperationParser


class Clip(Provider):
    model: str
    device: str | None

    _device: Any
    _model: Any
    _processor: Any

    def __init__(
        self,
        model: str = "openai/clip-vit-base-patch32",
        device: str | None = None,
        **kwargs,
    ):
        """Intialize.

        Args:
            model:
                CLIP model name.
            device:
                Pytorch device name.
        """
        self.model = model
        self.device = device

        self._model = None
        self._device = None
        self._processor = None

    def init(self, context: Context | None = None) -> None:
        if self._model:
            return
        self._device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )
        if self.device is not None:
            self._device = self.device
        self._model = CLIPModel.from_pretrained(self.model).to(self._device)
        self._processor = CLIPProcessor.from_pretrained(self.model)

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.init(context=context)
        op_parser = ZeroShotImageClassificationOperationParser(operation)
        # CLASSIFY
        if op_parser.op_equals(ZeroShotImageClassificationOperation.CLASSIFY):
            inputs = self._processor(
                images=op_parser.get_image_as_pil(),
                text=op_parser.get_classes(),
                return_tensors="pt",
                padding=True,
            ).to(self._device)
            with torch.no_grad():
                outputs = self._model(**inputs)
            logits_per_image = outputs.logits_per_image
            probs = logits_per_image.softmax(dim=1)
            result = probs.detach().numpy().tolist()[0]
        # BATCH
        elif op_parser.op_equals(ZeroShotImageClassificationOperation.BATCH):
            inputs = self._processor(
                images=op_parser.get_images_as_pil(),
                text=op_parser.get_classes(),
                return_tensors="pt",
                padding=True,
            ).to(self._device)
            with torch.no_grad():
                outputs = self._model(**inputs)
            logits_per_image = outputs.logits_per_image
            probs = logits_per_image.softmax(dim=1)
            result = probs.detach().numpy().tolist()
        else:
            return super().run(
                operation,
                context,
                **kwargs,
            )
        return Response(result=result)



from verse.core import DataModel
from typing import Any

class SpeechRecognitionResult(DataModel):
    """Speech recognition result."""

    text: str | None
    language: str | None
    verbose: Any | None
    """Transcribed text."""



from typing import Any, Callable, Generic, Literal, TypeVar

from verse.content.image import Image, ImageParam
from verse.core import DataModel
from verse.core.data_model import DataModelField


class ChatCompletionMessageContentPart(DataModel):
    type: Literal["text", "image"]
    """Content part type."""

    text: str | None = None
    """Text for text content type."""

    image: Image | ImageParam | dict | None = None
    """Image for image content type."""


class ToolCallFunction(DataModel):
    name: str
    """The name of the function to call."""

    arguments: str | None = None
    """The arguments to call the function in JSON format."""


class ToolCall(DataModel):
    id: str | None = None
    """The ID of the tool call."""

    function: ToolCallFunction
    """The function that the model called."""

    type: Literal["function"] = "function"
    """The type of the tool. Currently, only `function` is supported."""


class ChatCompletionMessage(DataModel):
    role: Literal["system", "user", "assistant", "tool"] | None = None
    """The role of the author."""

    content: str | list[ChatCompletionMessageContentPart] | None = None
    """The contents of the message."""

    name: str | None = None
    """An optional name for the participant."""

    tool_calls: list[ToolCall] | None = None
    """The tool calls generated by the model."""

    tool_call_id: str | None = None
    """Tool call that this message is responding to."""

    refusal: str | None = None
    """The refusal message generated by the model."""


class FunctionDefinition(DataModel):
    name: str
    """Name of the function."""

    description: str | None = None
    """Description of the function."""

    parameters: dict[str, Any] | None = None
    """Parameters of the function."""

    strict: bool | None = None
    """Whether to enable strict schema adherence
    when generating the function call.
    """


class Tool(DataModel):
    function: FunctionDefinition
    """Function definition."""

    type: Literal["function"] = "function"


class ToolChoiceFunction(DataModel):
    name: str
    """Name of the function."""


class ToolChoice(DataModel):
    function: ToolChoiceFunction

    type: Literal["function"] = "function"


class JSONSchema(DataModel):
    name: str
    """The name of the response format."""

    description: str
    """A description of what the response format is for."""

    schema_: dict[str, object] = DataModelField(alias="schema")
    """The schema for the response format, as a JSON Schema object."""

    strict: bool | None = None
    """Whether to enable strict schema adherence when generating the output."""


class ResponseFormat(DataModel):
    type: Literal["text", "json_object", "json_schema"]
    """Response format type."""

    json_schema: str


class StreamOptions(DataModel):
    include_usage: bool
    """A value that indicates if usage info should
    be sent as part of streaming.
    """


class TopLogprob(DataModel):
    token: str
    """The token."""

    bytes: list[int] | None = None
    """A list of integers representing the UTF-8
    bytes representation of the token.
    """

    logprob: float
    """The log probability of this token.
    """


class ChatCompletionTokenLogprob(DataModel):
    token: str
    """The token."""

    bytes: list[int] | None = None
    """A list of integers representing the UTF-8
    bytes representation of the token.
    """

    logprob: float
    """The log probability of this token.
    """

    top_logprobs: list[TopLogprob]
    """List of the most likely tokens and their log probability, at this token
    position.
    """


class ChoiceLogProbs(DataModel):
    content: list[ChatCompletionTokenLogprob] | None = None
    """A list of message content tokens with
    log probability information.
    """


class Choice(DataModel):
    finish_reason: str | None = None
    """The reason the model stopped generating tokens.
    Values are ["stop", "length", "content_filter", "tool_calls",
    "error", "cancel"]
    """

    index: int | None = None
    """The index of the choice in the list of choices."""

    message: ChatCompletionMessage | None = None
    """Chat completion response generated by the model."""

    delta: ChatCompletionMessage | None = None
    """Delta message when streaming."""

    logprobs: ChoiceLogProbs | None = None
    """Log probability information for the choice."""


class Usage(DataModel):
    completion_tokens: int
    """Number of tokens in the generated completion."""

    prompt_tokens: int
    """Number of tokens in the prompt."""

    total_tokens: int
    """Total number of tokens used in the request (prompt + completion)."""


class ChatCompletionResult(DataModel):
    id: str | None = None
    """A unique identifier for the chat completion."""

    choices: list[Choice]
    """A list of chat completion choices."""

    created: int | None = None
    """The Unix timestampof when the chat completion was created."""

    model: str | None = None
    """The model used for the chat completion."""

    usage: Usage | None = None
    """Usage statistics for the completion request."""


T = TypeVar("T")


class Stream(Generic[T]):
    stream: Any
    convert_chunk: Callable[[Any], T]

    def __init__(self, stream: Any, convert_chunk: Callable[[Any], T]):
        self.stream = stream
        self.convert_chunk = convert_chunk

    def __iter__(self):
        return self

    def __next__(self) -> T:
        try:
            chunk = self.stream.__next__()
            return self.convert_chunk(chunk)
        except StopAsyncIteration:
            raise

    def __aiter__(self):
        return self

    async def __anext__(self) -> T:
        try:
            chunk = await self.stream.__anext__()
            return self.convert_chunk(chunk)
        except StopAsyncIteration:
            raise



from typing import Any

from verse.content.audio import Audio, AudioParam
from verse.core import Component, Response

from ._models import SpeechRecognitionResult
from ._operation import SpeechRecognitionOperation


class SpeechRecognition(Component):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def transcribe(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[SpeechRecognitionResult]:
        """Transcribe audio and convert to text.

        Args:
            audio: Audio to transcibe.

        Returns:
            Transciption result.
        """
        return self._run_internal(
            SpeechRecognitionOperation.TRANSCRIBE, locals()
        )

    def detect_language(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[SpeechRecognitionResult]:
        """Detect langauge in audio.

        Args:
            audio: Audio to recognize.

        Returns:
            Detection result.
        """
        return self._run_internal(
            SpeechRecognitionOperation.DETECT_LANGUAGE, locals()
        )

    async def atranscribe(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[SpeechRecognitionResult]:
        """Transcribe audio and convert to text.

        Args:
            audio: Audio to transcibe.

        Returns:
            Transciption result.
        """
        return await self._arun_internal(
            SpeechRecognitionOperation.TRANSCRIBE, locals()
        )

    async def adetect_language(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[SpeechRecognitionResult]:
        """Detect langauge in audio.

        Args:
            audio: Audio to recognize.

        Returns:
            Detection result.
        """
        return await self._arun_internal(
            SpeechRecognitionOperation.DETECT_LANGUAGE, locals()
        )



from __future__ import annotations

__all__ = ["OpenAI"]

from typing import Any

from verse.content.image import Image, ImageParam
from verse.core import Context, NCall, Operation, Provider, Response
from verse.core.exceptions import BadRequestError

from .._helper import copy_arg
from .._models import ChatCompletionResult, Stream
from .._operation import ChatCompletionOperation
from .._operation_parser import ChatCompletionOperationParser


class OpenAI(Provider):
    model: str | None
    api_key: str | None
    organization: str | None
    project: str | None
    base_url: str | None
    timeout: float | None
    service_tier: str | None
    nparams: dict

    _client: Any
    _aclient: Any

    def __init__(
        self,
        model: str | None = "gpt-4o",
        api_key: str | None = None,
        organization: str | None = None,
        project: str | None = None,
        base_url: str | None = None,
        timeout: float | None = None,
        service_tier: str | None = None,
        nparams: dict = dict(),
    ):
        """Initialize.

        Args:
            model:
                OpenAI model to use for chat completion.
            api_key:
                OpenAI API key.
            organization:
                OpenAI organization.
            project:
                OpenAI project.
            base_url:
                OpenAI base url.
            timeout:
                Timeout for client.
            service_tier:
                OpenAI service tier.
            nparams:
                Native params for OpenAI client.
        """
        self.model = model
        self.api_key = api_key
        self.organization = organization
        self.project = project
        self.base_url = base_url
        self.timeout = timeout
        self.service_tier = service_tier
        self.nparams = nparams

        self._client = None
        self._aclient = None

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        from openai import OpenAI

        self._client = OpenAI(**self._get_client_params())

    async def ainit(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        from openai import AsyncOpenAI

        self._aclient = AsyncOpenAI(**self._get_client_params())

    def _get_client_params(self) -> dict[str, Any]:
        args: dict = {}
        if self.api_key is not None:
            args["api_key"] = self.api_key
        if self.organization is not None:
            args["organization"] = self.organization
        if self.project is not None:
            args["project"] = self.project
        if self.base_url is not None:
            args["base_url"] = self.base_url
        if self.timeout is not None:
            args["timeout"] = self.timeout
        args = args | self.nparams
        return args

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.init(context=context)
        op_parser = ChatCompletionOperationParser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            self._client,
        )
        if ncall is None:
            return super().run(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(call=ncall, result=nresult))

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.ainit(context=context)
        op_parser = ChatCompletionOperationParser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            self._aclient,
        )
        if ncall is None:
            return await super().arun(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(call=ncall, result=nresult))

    def _get_ncall(
        self,
        op_parser: ChatCompletionOperationParser,
        client: Any,
    ) -> tuple[NCall | None, dict]:
        call = None
        state: dict = {}
        nargs = op_parser.get_nargs()
        op_converter = OperationConverter(self.model, self.service_tier)
        # COMPLETE
        if op_parser.op_equals(ChatCompletionOperation.COMPLETE):
            args = op_converter.convert_complete(op_parser.get_args())
            call = NCall(
                client.chat.completions.create,
                args,
                nargs,
            )
        # STREAM
        elif op_parser.op_equals(ChatCompletionOperation.STREAM):
            args = op_converter.convert_stream(op_parser.get_args())
            call = NCall(
                client.chat.completions.create,
                args,
                nargs,
            )
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict,
        op_parser: ChatCompletionOperationParser,
    ) -> Any:
        result: Any = None
        result_converter = ResultConverter()
        # COMPLETE
        if op_parser.op_equals(ChatCompletionOperation.COMPLETE):
            result = result_converter.convert_complete_result(
                nresult,
            )
        # STREAM
        if op_parser.op_equals(ChatCompletionOperation.STREAM):
            result = result_converter.convert_stream_result(
                nresult,
            )
        return result


class ResultConverter:
    def __init__(self):
        pass

    def convert_complete_result(
        self,
        nresult: Any,
    ) -> ChatCompletionResult:
        return ChatCompletionResult.model_validate(nresult.model_dump())

    def convert_stream_result(
        self,
        nresult: Any,
    ) -> Stream:
        return Stream(nresult, self.convert_complete_result)


class OperationConverter:
    model: str | None
    service_tier: str | None

    def __init__(self, model: str | None, service_tier: str | None):
        self.model = model
        self.service_tier = service_tier

    def convert_complete(
        self,
        args: dict,
    ) -> dict:
        return self._convert_args(args)

    def convert_stream(
        self,
        args: dict,
    ) -> dict:
        rargs = self._convert_args(args)
        rargs["stream"] = True
        return rargs

    def _convert_messages(self, messages: list):
        for message in messages:
            content = message["content"]
            if isinstance(content, list):
                for part in content:
                    if part["type"] == "image":
                        image_part = part["image"]
                        if isinstance(image_part, Image):
                            image = image_part
                        elif isinstance(image_part, ImageParam):
                            image = Image(param=image_part)
                        elif isinstance(image_part, dict):
                            image = Image(
                                param=ImageParam.from_dict(image_part)
                            )
                        else:
                            raise BadRequestError("Image content not valid.")
                        if image.url is not None:
                            part["image_url"] = {"url": image.url}
                        else:
                            base64 = image.convert(
                                type="base64", format="jpeg"
                            )
                            part["image_url"] = {
                                "url": f"data:image/jpeg;base64,{base64}"
                            }
                        part.pop("image")
                        part["type"] = "image_url"

    def _convert_args(self, args: dict) -> dict:
        rargs: dict = {}
        if "model" not in args:
            rargs["model"] = self.model
        else:
            copy_arg(args, rargs, "model")
        if self.service_tier is not None:
            args["service_tier"] = self.service_tier
        copy_arg(args, rargs, "messages")
        self._convert_messages(rargs["messages"])
        copy_arg(args, rargs, "temperature")
        copy_arg(args, rargs, "top_p")
        copy_arg(args, rargs, "max_tokens")
        copy_arg(args, rargs, "stop")
        copy_arg(args, rargs, "n")
        copy_arg(args, rargs, "presence_penalty")
        copy_arg(args, rargs, "frequency_penalty")
        copy_arg(args, rargs, "tools")
        copy_arg(args, rargs, "tool_choice")
        copy_arg(args, rargs, "parallel_tool_calls")
        copy_arg(args, rargs, "response_format")
        copy_arg(args, rargs, "seed")
        copy_arg(args, rargs, "logit_bias")
        copy_arg(args, rargs, "logprobs")
        copy_arg(args, rargs, "stream_options")
        copy_arg(args, rargs, "top_logprobs")
        copy_arg(args, rargs, "user")
        copy_arg(args, rargs, "extra_headers")
        copy_arg(args, rargs, "extra_query")
        copy_arg(args, rargs, "extra_body")
        copy_arg(args, rargs, "timeout")
        return rargs



from __future__ import annotations

from typing import Any

from verse.core import DataModel, OperationParser

from ._models import (
    ChatCompletionMessage,
    ResponseFormat,
    StreamOptions,
    Tool,
    ToolChoice,
)


class ChatCompletionOperationParser(OperationParser):
    def get_args(self, strong_types: bool = False) -> dict:
        args = super().get_args()
        if "messages" in args:
            if isinstance(args["messages"], str):
                args["messages"] = [
                    {
                        "role": "user",
                        "content": args["messages"],
                    }
                ]

        if not strong_types:
            for name, value in args.items():
                if isinstance(value, DataModel):
                    args[name] = value.to_dict()
                if isinstance(value, list):
                    for i, _ in enumerate(value):
                        if isinstance(value[i], DataModel):
                            value[i] = value[i].to_dict()
            return args

        types_map: dict[str, Any] = {
            "messages": ChatCompletionMessage,
            "stream_options": StreamOptions,
            "tools": Tool,
            "tool_choice": ToolChoice,
            "response_format": ResponseFormat,
        }
        for name, value in args.items():
            if name in types_map:
                if isinstance(value, list):
                    for i, _ in enumerate(value):
                        if isinstance(value[i], dict):
                            value[i] = types_map[name].from_dict(value[i])
                else:
                    if isinstance(value[i], dict):
                        args[name] = types_map[name].from_dict(value)
        return args



"""
Image captioning using Chat Completion Provider.
"""

__all__ = ["ChatCompletionProvider"]

import json
from typing import Any

from verse.ai.chat_completion._operation import ChatCompletionOperation
from verse.content.image import Image
from verse.core import Component, Context, Operation, Provider, Response

from .._operation import ImageCaptioningOperation
from .._operation_parser import ImageCaptioningOperationParser


class ChatCompletionProvider(Provider):
    component: Component

    def __init__(
        self,
        component: Component,
        **kwargs,
    ):
        """Intialize.

        Args:
            component:
                Chat completion component.
        """
        self.component = component

    def init(self, context: Context | None = None) -> None:
        self.component.init(context=context)

    async def ainit(self, context: Context | None = None) -> None:
        await self.component.ainit(context=context)

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.init(context=context)
        result = None
        op_parser = ImageCaptioningOperationParser(operation)
        noperation, state = self._get_operation(op_parser)
        if noperation is not None:
            nresult = self.component.run(
                noperation,
                context,
                **kwargs,
            )
            result = self._convert_result(op_parser, nresult, state)
            return Response(result=result, native=dict(result=nresult))
        else:
            return super().run(
                operation,
                context,
                **kwargs,
            )

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.ainit(context=context)
        result = None
        op_parser = ImageCaptioningOperationParser(operation)
        noperation, state = self._get_operation(op_parser)
        if noperation is not None:
            nresult = await self.component.arun(
                noperation,
                context,
                **kwargs,
            )
            result = self._convert_result(op_parser, nresult, state)
            return Response(result=result, native=dict(result=result))
        else:
            return super().run(
                operation,
                context,
                **kwargs,
            )

    def _get_operation(
        self, op_parser: ImageCaptioningOperationParser
    ) -> tuple[Operation, dict]:
        state = {}
        converter = OperationConverter()
        # GENERATE
        if op_parser.op_equals(ImageCaptioningOperation.GENERATE):
            state["n"] = 1
            operation = converter.convert_generate(
                op_parser.get_image(), op_parser.get_max_length()
            )
        # BATCH
        elif op_parser.op_equals(ImageCaptioningOperation.BATCH):
            images = op_parser.get_images()
            state["n"] = len(images)
            operation = converter.convert_batch(
                images, op_parser.get_max_length()
            )
        else:
            return None, None
        return operation, state

    def _convert_result(
        self,
        op_parser: ImageCaptioningOperationParser,
        nresult: Any,
        state: dict,
    ) -> Any:
        nresult = nresult.result
        converter = ResultConverter()
        result: Any = None
        # GENERATE
        if op_parser.op_equals(ImageCaptioningOperation.GENERATE):
            result = converter.convert_generate_result(nresult)
        # BATCH
        elif op_parser.op_equals(ImageCaptioningOperation.BATCH):
            n = state["n"]
            result = converter.convert_batch_result(nresult, n)
        return result


class ResultConverter:
    def convert_generate_result(self, nresult: Any) -> str:
        return nresult.choices[0].message.content

    def convert_batch_result(self, result: Any, n: int) -> list[str]:
        json_string = result.choices[0].message.content
        return json.loads(json_string)["captions"]


class OperationConverter:
    def convert_generate(
        self,
        image: Image,
        max_length: int | None = None,
        **kwargs,
    ) -> Operation:
        system_content = "You are an image captioning provider that captions images. For the given image provide a caption in plain text."  # noqa
        if max_length is not None:
            system_content = (
                system_content
                + f" The maximum number of tokens in the caption cannot exceed {max_length}."  # noqa
            )
        messages = [
            {
                "role": "system",
                "content": system_content,
            },
            {
                "role": "user",
                "content": [
                    {
                        "type": "image",
                        "image": image,
                    }
                ],
            },
        ]
        return Operation(
            name=ChatCompletionOperation.COMPLETE,
            args={"messages": messages},
        )

    def convert_batch(
        self,
        images: list[Image],
        max_length: int | None = None,
        **kwargs,
    ) -> Operation:
        system_content = """You are an image captioning provider that captions a batch of images. For each image, provide a caption in plain text. One image is not related to another. Return the batch of captions as a JSON object with one array field called 'captions' where each value in the array is the caption of the corresponding image."""  # noqa
        if max_length is not None:
            system_content = (
                system_content
                + f" The maximum number of tokens in the caption cannot exceed {max_length}."  # noqa
            )
        content_parts = []
        for image in images:
            content_parts.append(
                {
                    "type": "image",
                    "image": image,
                }
            )
        messages = [
            {
                "role": "system",
                "content": system_content,
            },
            {"role": "user", "content": content_parts},
        ]
        return Operation(
            name=ChatCompletionOperation.COMPLETE,
            args={
                "messages": messages,
                "response_format": {"type": "json_object"},
            },
        )



def copy_arg(
    source: dict, dest: dict, source_name: str, dest_name: str | None = None
):
    if source_name in source:
        if dest_name is None:
            dest[source_name] = source[source_name]
        else:
            dest[dest_name] = source[source_name]
